#include <THC/THCNumerics.cuh>
#include <THC/THCReduceApplyUtils.cuh>
#include <THCUNN/SharedMem.cuh>
#include <ATen/ATen.h>
#include <ATen/cuda/CUDAContext.h>
#include <THC/THC.h>
#include <THC/THCAtomics.cuh>
#include <THC/THCDeviceUtils.cuh>

#define CHECK_CUDA(x) AT_ASSERTM(x.type().is_cuda(), #x " must be a CUDA tensor")

const int WARP_SIZE = 32;
// Crude benchmarks suggest 256 is better than 512 and 1024
// TODO: Autotune/use better heuristics, improve speed more.
const int MAX_BLOCK_SIZE = 256;
const int CUDA_NUM_THREADS = 1024;

int GET_BLOCKS(const int N) {
  return (N + CUDA_NUM_THREADS - 1) / CUDA_NUM_THREADS;
}

static int getGradParamsNumThreads(int batchSize){
//warp per item in a batch, up to a maximum
   return std::min(batchSize * WARP_SIZE, MAX_BLOCK_SIZE);

}

template <typename T>
struct ReduceAdd {
  inline __device__ T operator()(const T a, const T b) const {
    return THCNumerics<T>::add(a, b);
  }
};

template <typename T, int NBoxes, int PPB>
__global__ void spatialDepthwiseBoxFilterConvolutionUpdateOutput(
    const T* input_data,
    T* output_data,
    const T* weight_data,
    const T* sampleLoc_data,
    const T* bias_data,
    bool biasEnabled,
    const int totalElements,
    const int outputChannels,
    const int inputWidth, const int inputHeight,
    const int outputWidth, const int outputHeight,
    const int strideWidth, const int strideHeight,
    const int padWidth, const int padHeight,
    const int numBoxes, const int pixelPerBox)
{
  const int BOXES_LIMIT = (NBoxes !=0) ? NBoxes : numBoxes;
  const int PIXEL_PER_BOX = (PPB != 0) ? PPB : pixelPerBox;

  for (int linearIndex = blockIdx.x * blockDim.x + threadIdx.x;
       linearIndex < totalElements;
       linearIndex += gridDim.x * blockDim.x) {

    int indtmp1 = linearIndex/outputWidth;
    const int w = linearIndex - indtmp1 * outputWidth;
    int indtmp2 = indtmp1/outputHeight;
    const int h = indtmp1 - indtmp2 * outputHeight;
    indtmp1 = indtmp2;
    indtmp2 = indtmp1/outputChannels;
    const int c = indtmp1 - indtmp2 * outputChannels;
    const int n = indtmp2;

    int inputChannel = c;
    int inputChannels = outputChannels;

    int weightOffset = c * BOXES_LIMIT * PIXEL_PER_BOX;
    int sampleLocOffset = weightOffset * 2;

    T value = biasEnabled ? bias_data[c] : T(0);
    const int offset0 = (n * inputChannels + inputChannel) * 
                        inputHeight * inputWidth;
#ifndef __HIP_PLATFORM_HCC__
#pragma unroll
#endif
    for (int boxIdx = 0; boxIdx < BOXES_LIMIT; ++boxIdx) {
#ifndef __HIP_PLATFORM_HCC__
#pragma unroll
#endif
      for (int sampleIdx = 0; sampleIdx < PIXEL_PER_BOX; ++sampleIdx) {
        const int w_in = min(-padWidth + w * strideWidth +
                          (int)sampleLoc_data[sampleLocOffset++], inputWidth - 1);
        const int h_in = min(-padHeight + h * strideHeight + 
                          (int)sampleLoc_data[sampleLocOffset++], inputHeight - 1);

        if ((h_in >= 0) && (h_in < inputHeight) && (w_in >= 0) && (w_in < inputWidth)) {
          const int offset = offset0 + h_in * inputWidth + w_in;
          value += weight_data[weightOffset] * input_data[offset];
        }
        ++weightOffset;
      }
    }
    output_data[linearIndex] = value;
  }
}


template <typename T, int NBoxes, int PPB, int stride>
__global__ void spatialDepthwiseBoxFilterConvolutionUpdateGradInput(
    const T* gradOutput_data,
    T* gradInput_data,
    const T* weight_data,
    const T* sampleLoc_data,
    const int totalElements,
    const int outputChannels,
    const int inputWidth, const int inputHeight,
    const int outputWidth, const int outputHeight,
    const int strideWidth, const int strideHeight,
    const int padWidth, const int padHeight,
    const int numBoxes, const int pixelPerBox)
{
  const int BOXES_LIMIT = (NBoxes !=0) ? NBoxes : numBoxes;
  const int PIXEL_PER_BOX = (PPB != 0) ? PPB : pixelPerBox;
  const int strideW = (stride !=0) ? stride : strideWidth;
  const int strideH = (stride !=0) ? stride : strideHeight;
  const int inputWidthActual = inputWidth - padWidth;
  const int inputHeightActual = inputHeight - padHeight;
  
  for (int linearIndex = blockIdx.x * blockDim.x + threadIdx.x;
       linearIndex < totalElements;
       linearIndex += gridDim.x * blockDim.x) {

    int indtmp1 = linearIndex/inputWidth;
    const int w = linearIndex - indtmp1 * inputWidth;
    int indtmp2 = indtmp1/inputHeight;
    const int h = indtmp1 - indtmp2 * inputHeight;
    indtmp1 = indtmp2;
    indtmp2 = indtmp1/outputChannels;
    const int c = indtmp1 - indtmp2 * outputChannels;
    const int n = indtmp2;


    int och = c;
    int weightOffset = c * BOXES_LIMIT * PIXEL_PER_BOX;
    int sampleLocOffset = weightOffset * 2;

    T value(0);

#ifndef __HIP_PLATFORM_HCC__
#pragma unroll
#endif
    for (int boxIdx = 0; boxIdx < BOXES_LIMIT; ++boxIdx) {
#ifdef __HIP_PLATFORM_HCC__
#pragma unroll
#endif
      for (int sampleIdx = 0; sampleIdx < PIXEL_PER_BOX; ++sampleIdx) {
        int w_out = w + padWidth - sampleLoc_data[sampleLocOffset++];
        int h_out = h + padHeight - sampleLoc_data[sampleLocOffset++];
        if ((h_out % strideH == 0) && (w_out % strideW == 0)) {
          h_out = h_out / strideH;
          w_out = w_out / strideW;

          if ((h_out >= 0) && (h_out < outputHeight)
                && (w_out >= 0) && (w_out < outputWidth)) {

            const int offset = ((n * outputChannels + och) * outputHeight + h_out)
                    * outputWidth + w_out;
            value += weight_data[weightOffset] * gradOutput_data[offset];
          }
        }
        ++weightOffset;
      }
    }

    const int w_sample = min(w, inputWidthActual - 1);
    const int h_sample = min(h, inputHeightActual - 1);
    if (w_sample == inputWidthActual - 1 || h_sample == inputHeightActual - 1) {
      atomicAdd(&gradInput_data[((n * outputChannels + c) * inputHeightActual + h_sample) * inputWidthActual + w_sample], value);
    } else {
      gradInput_data[((n * outputChannels + c) * inputHeightActual + h_sample) * inputWidthActual + w_sample] = value;
    }
  }
}

template <typename T>
__global__ void spatialDepthwiseBoxFilterConvolutionAccGradParameters(
    const T* gradOutput_data,
    const T* input_data,
    T* gradWeight_data,
    const T* sampleLoc_data,
    const int batchSize,
    const int numChannels,
    const int inputWidth, const int inputHeight,
    const int outputWidth, const int outputHeight,
    const int strideWidth, const int strideHeight,
    const int padWidth, const int padHeight,
    const int numBoxes, const int pixelPerBox)
{
  const int channelStride = numBoxes * pixelPerBox;

  // Have to use a statically typed Shared Memory pointer
  SharedMem<T> smem;

  // Each Block is responsible for accumulating over a permutation of
  // (channels x kH x kW), use blockIdx to determine which one
  int bidx = blockIdx.x;
  int sampleIdx = bidx % pixelPerBox;
  int boxIdx = (bidx / pixelPerBox) % numBoxes;
  int ch = (bidx / channelStride);

  int sampleLocOffset = ((ch * numBoxes + boxIdx) * pixelPerBox + sampleIdx) * 2;
  int kx = sampleLoc_data[sampleLocOffset];
  int ky = sampleLoc_data[sampleLocOffset + 1];

  // Need to calculate which input channel is associated with this filter
  // channel
  int inputCh = ch;

  T grad(0);

  const int laneId = threadIdx.x % WARP_SIZE;
  const int batch = threadIdx.x / WARP_SIZE;
  const int nwarps = blockDim.x / WARP_SIZE;
  const int imageElements = outputWidth * outputHeight;
  // Use warp per item.  In the original kernel, a threadblock was used to sum over NHW.
  // Here, we use a warp to sum values over HW dimension, and if batchSize is larger than the
  // number of warps, a warp would loop over remaining batch items (e.g. if there are 8 warps,
  // warp 0 would go over 0-8-16 etc image, warp 1 over 1-9-17 etc). Later in blockReduce,
  // all the warps will be reduced anyway, thus the full reduction will be over NHW, like it
  // should be. That allows to get rid of one modulo operation inside the loop (because n/batchIdx
  // now does not have to be computed through modulo, you are just looping over it), and
  // bring a nice speed-up.
  for (int batchIdx = batch; batchIdx < batchSize; batchIdx += nwarps){
    // Warp-stride loop over elements in a batch item
    for (int idx = laneId; idx < imageElements; idx += WARP_SIZE) {
    // Need to calculate the following: batch position, and offset into the gradOutput
    // in height, and width. We can intuit the corresponding position in the input from
    // the other parameters we have
      int go_w_offset = idx % outputWidth;
      int go_h_offset = (idx / outputWidth);

      // pay attention to right and bottom boundaries of summed area table
      int i_w_offset = min((go_w_offset * strideWidth) + kx - padWidth, inputWidth - 1);
      int i_h_offset = min((go_h_offset * strideHeight) + ky - padHeight, inputHeight - 1);

      if (i_w_offset >= 0 && i_h_offset >= 0 && i_w_offset < inputWidth && i_h_offset < inputHeight) {
        int inputOffset = ((batchIdx * numChannels + inputCh) * inputHeight + i_h_offset) * inputWidth + i_w_offset;
        int outputOffset = ((batchIdx * numChannels + ch) * outputHeight ) * outputWidth + idx;
        grad += input_data[inputOffset] * gradOutput_data[outputOffset];
      }
    }
  }
  __syncthreads();

  // At this point each thread in the block has a local gradient, which we need to
  // accumulate prior to writing the global value
  T *buf = smem.getPointer();
  T tval = reduceBlock<T, ReduceAdd<T>>(
      buf, blockDim.x, grad, ReduceAdd<T>(), T(0));

  // After reduction, first thread in the block has the gradient, so its responsible
  // for writing it to gradWeight
  if (threadIdx.x == 0) {
    int weightOffset = sampleIdx + (pixelPerBox * boxIdx) + (channelStride * ch);
    gradWeight_data[weightOffset] = tval;
  }
}



void SpatialDepthwiseBoxFilterConvolution_forward_cuda(
  const at::Tensor& SAT,
  const at::Tensor& sampleLoc,
  const at::Tensor& weight,
  const at::Tensor& bias,
  const int ksize,
  const int stride,
  const int pad,
  at::Tensor& output)
{
  AT_ASSERTM(SAT.type().is_cuda(), "SAT must be a CUDA tensor");
  AT_ASSERTM(sampleLoc.type().is_cuda(), "sampleLoc must be a CUDA tensor");
  AT_ASSERTM(weight.type().is_cuda(), "weights must be a CUDA tensor");
  AT_ASSERTM(output.type().is_cuda(), "output must be a CUDA tensor");

  cudaStream_t stream = at::cuda::getCurrentCUDAStream();

  int batchSize    = SAT.size(0);
  int inputHeight  = SAT.size(2);
  int inputWidth   = SAT.size(3);
  int outputHeight = output.size(2);
  int outputWidth  = output.size(3);
  int numChannels  = weight.size(0);
  int numBoxes     = weight.size(1);
  int pixelPerBox  = weight.size(2);

  // One thread per output value
  int nthreads = output.numel();
  int blocks = GET_BLOCKS(nthreads);
  dim3 grid(blocks);
  dim3 block(CUDA_NUM_THREADS);

  if (numBoxes == 1 && pixelPerBox == 16) {
    AT_DISPATCH_FLOATING_TYPES(SAT.type(), "SpatialDepthwiseBoxFilterConvolution_forward_cuda", [&] {
      spatialDepthwiseBoxFilterConvolutionUpdateOutput<scalar_t, 1, 16><<<grid, block, 0, stream>>>(
        SAT.data<scalar_t>(),
        output.data<scalar_t>(),
        weight.data<scalar_t>(),
        sampleLoc.data<scalar_t>(),
        bias.data<scalar_t>(),
        bias.numel() == 0 ? false : true,
        nthreads,
        numChannels,
        inputWidth, inputHeight,
        outputWidth, outputHeight,
        stride, stride,
        pad, pad,
        numBoxes, pixelPerBox
      );
    });
  } else if (numBoxes == 1 && pixelPerBox == 4) {
    AT_DISPATCH_FLOATING_TYPES(SAT.type(), "SpatialDepthwiseBoxFilterConvolution_forward_cuda", [&] {
      spatialDepthwiseBoxFilterConvolutionUpdateOutput<scalar_t, 1, 4><<<grid, block, 0, stream>>>(
        SAT.data<scalar_t>(),
        output.data<scalar_t>(),
        weight.data<scalar_t>(),
        sampleLoc.data<scalar_t>(),
        bias.data<scalar_t>(),
        bias.numel() == 0 ? false : true,
        nthreads,
        numChannels,
        inputWidth, inputHeight,
        outputWidth, outputHeight,
        stride, stride,
        pad, pad,
        numBoxes, pixelPerBox
      );
    });
  } else {
    AT_DISPATCH_FLOATING_TYPES(SAT.type(), "SpatialDepthwiseBoxFilterConvolution_forward_cuda", [&] {
      spatialDepthwiseBoxFilterConvolutionUpdateOutput<scalar_t, 0, 0><<<grid, block, 0, stream>>>(
        SAT.data<scalar_t>(),
        output.data<scalar_t>(),
        weight.data<scalar_t>(),
        sampleLoc.data<scalar_t>(),
        bias.data<scalar_t>(),
        bias.numel() == 0 ? false : true,
        nthreads,
        numChannels,
        inputWidth, inputHeight,
        outputWidth, outputHeight,
        stride, stride,
        pad, pad,
        numBoxes, pixelPerBox
      );
    });
  }
  THCudaCheck(cudaGetLastError());
}


void SpatialDepthwiseBoxFilterConvolution_backward_cuda(
  const at::Tensor& SAT,
  const at::Tensor& sampleLoc,
  const at::Tensor& weight,
  const at::Tensor& grad_output,
  const int ksize,
  const int stride,
  const int pad,
  at::Tensor& grad_SAT,
  at::Tensor& grad_weight) 
{
  CHECK_CUDA(SAT);
  CHECK_CUDA(sampleLoc);
  CHECK_CUDA(weight);
  CHECK_CUDA(grad_output);
  CHECK_CUDA(grad_SAT);
  CHECK_CUDA(grad_weight);

  cudaStream_t stream = at::cuda::getCurrentCUDAStream();

  int batchSize    = SAT.size(0);
  int inputHeight  = SAT.size(2);
  int inputWidth   = SAT.size(3);
  int outputHeight = grad_output.size(2);
  int outputWidth  = grad_output.size(3);
  int numChannels  = weight.size(0);
  int numBoxes     = weight.size(1);
  int pixelPerBox  = weight.size(2);

  // int nthreads = grad_SAT.numel();
  int nthreads = batchSize * numChannels * (inputHeight + pad) * (inputWidth + pad);
  int blocks = GET_BLOCKS(nthreads);
  dim3 grid(blocks);
  dim3 block(CUDA_NUM_THREADS);

  if (numBoxes == 1 && pixelPerBox == 16 && stride == 1) {
    AT_DISPATCH_FLOATING_TYPES(grad_SAT.type(), "SpatialDepthwiseBoxFilterConvolution_backward_cuda", [&] {
      spatialDepthwiseBoxFilterConvolutionUpdateGradInput<scalar_t, 1, 16, 1><<<grid, block, 0, stream>>>(
        grad_output.data<scalar_t>(),
        grad_SAT.data<scalar_t>(),
        weight.data<scalar_t>(),
        sampleLoc.data<scalar_t>(),
        nthreads,
        numChannels,
        inputWidth + pad, inputHeight + pad,
        outputWidth, outputHeight,
        stride, stride,
        pad, pad,
        numBoxes, pixelPerBox
      );
    });
  } else if (numBoxes == 1 && pixelPerBox == 4 && stride == 1) {
    AT_DISPATCH_FLOATING_TYPES(grad_SAT.type(), "SpatialDepthwiseBoxFilterConvolution_backward_cuda", [&] {
      spatialDepthwiseBoxFilterConvolutionUpdateGradInput<scalar_t, 1, 4, 1><<<grid, block, 0, stream>>>(
        grad_output.data<scalar_t>(),
        grad_SAT.data<scalar_t>(),
        weight.data<scalar_t>(),
        sampleLoc.data<scalar_t>(),
        nthreads,
        numChannels,
        inputWidth + pad, inputHeight + pad,
        outputWidth, outputHeight,
        stride, stride,
        pad, pad,
        numBoxes, pixelPerBox
      );
    });  
  } else {
    AT_DISPATCH_FLOATING_TYPES(grad_SAT.type(), "SpatialDepthwiseBoxFilterConvolution_backward_cuda", [&] {
      spatialDepthwiseBoxFilterConvolutionUpdateGradInput<scalar_t, 0, 0, 0><<<grid, block, 0, stream>>>(
        grad_output.data<scalar_t>(),
        grad_SAT.data<scalar_t>(),
        weight.data<scalar_t>(),
        sampleLoc.data<scalar_t>(),
        nthreads,
        numChannels,
        inputWidth + pad, inputHeight + pad,
        outputWidth, outputHeight,
        stride, stride,
        pad, pad,
        numBoxes, pixelPerBox
      );
    });
  }
  THCudaCheck(cudaGetLastError());


  blocks = numChannels * numBoxes * pixelPerBox;
  grid = dim3(blocks);
  block = dim3(getGradParamsNumThreads(batchSize));

  AT_DISPATCH_FLOATING_TYPES(grad_SAT.type(), "SpatialDepthwiseBoxFilterConvolution_backward_cuda", [&] {
    spatialDepthwiseBoxFilterConvolutionAccGradParameters<scalar_t>
      <<<grid, block, block.x * sizeof(scalar_t), stream>>>(
      grad_output.data<scalar_t>(),
      SAT.data<scalar_t>(),
      grad_weight.data<scalar_t>(),
      sampleLoc.data<scalar_t>(),
      batchSize,
      numChannels,
      inputWidth, inputHeight,
      outputWidth, outputHeight,
      stride, stride,
      pad, pad,
      numBoxes, pixelPerBox
    );
  });
  THCudaCheck(cudaGetLastError());
}
